//////////////////////////////////////////////////////////////////////////////
// ***** BEGIN LICENSE BLOCK *****
// Version: MPL 1.1/GPL 2.0/LGPL 2.1
//
// The contents of this file are subject to the Mozilla Public License Version
// 1.1 (the "License"); you may not use this file except in compliance with
// the License. You may obtain a copy of the License at
// http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// The Original Code is CamerAHBasis.pas .
//
// The Initial Developer of the Original Code is Jens Biermann.
// Portions created by the Initial Developer are Copyright (C) 2011
// the Initial Developer. All Rights Reserved.
//
// Contributor(s):
//
// Alternatively, the contents of this file may be used under the terms of
// either the GNU General Public License Version 2 or later (the "GPL"), or
// the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
// in which case the provisions of the GPL or the LGPL are applicable instead
// of those above. If you wish to allow use of your version of this file only
// under the terms of either the GPL or the LGPL, and not to allow others to
// use your version of this file under the terms of the MPL, indicate your
// decision by deleting the provisions above and replace them with the notice
// and other provisions required by the GPL or the LGPL. If you do not delete
// the provisions above, a recipient may use your version of this file under
// the terms of any one of the MPL, the GPL or the LGPL.
//
// ***** END LICENSE BLOCK *****
//////////////////////////////////////////////////////////////////////////////

unit CamerAHBasis;

interface

uses
  CamerAHGeometry, dglOpenGL, CamerAHTypes, Windows, Controls;

const
  crZoonIn    = 31;
  crZoomOut   = 32;
  crSelect    = 33;
  crRotate    = 34;
  crMove      = 35;
  crMoveDepth = 36;

  function RectWidth(const Rect: TRect): Integer;
  function RectHeight(const Rect: TRect): Integer;

type
  TRotMode = (rmFlexAxis, rmArcBall, rmMove, rmDepthMove,
              rmRotX, rmRotY, rmRotZ,
              rmRotCoordX, rmRotCoordY, rmRotCoordZ,
              rmZoomPlus, rmZoomMinus, rmNothing, rmSelect, rmColorPick);

  TView    = (XY, XZ, YX, YZ, ZX, ZY);

  TProjection = (pjOrtho, pjPerspectiv);

  TClipPlane = record
    left, right, bottom, top,
    znear, zfar: Single;
  end;

  TSelectionResults = array of Integer;

  TglCamerAHBasis = class(TObject)
  private
    FPointofRotation: TVector;
    FPointofOrgin: TVector;
    FPointofView: TVector;
    FPointofCamera: TVector;
    FIsMovePoR: Boolean;
    FXVectors: TVector;
    FYVectors: TVector;
    FZVectors: TVector;
    FIsScreenAdjust: Boolean;
    procedure SetPointofRotation(const Value: TVector);
    function  GetXVectors: TVector;
    function  GetYVectors: TVector;
    function  GetZVectors: TVector;
    procedure SetPointofOrgin(const Value: TVector);
    procedure SetPointofView(const Value: TVector);
    procedure SetPointofCamera(const Value: TVector);
    procedure SetIsMovePoR(const Value: Boolean);
    procedure SetIsScreenAdjust(const Value: Boolean);
  protected
    tmpProjMatrix: TMatrix;
    EnableRotion: Boolean;
    FRotMode: TRotMode;
    ProjectionMode: TProjection;
    xStart,
    yStart: Integer;
    SelectBuffer : array[0..512] of TGLUInt;

    procedure CM_Multiply(M: TMatrix);
    procedure Colorpicking(X, Y: Integer; var R, G, B, A: GLubyte); overload;
    procedure Colorpicking(X, Y: Integer); overload;
    function  DistanceCameraCenter: Single;
    function  DistanceCameraRotation: Single;
    function  MoveVector(orientation: TVector; PixelDis: Integer): TVector;
    function  MoveVectorX(PixelDis: Integer): TVector;
    function  MoveVectorY(PixelDis: Integer): TVector;
    function  MoveVectorZ(PixelDis: Integer): TVector;
    function  OGLPos(AV: TVector): TVector; overload;
    function  OGLPos(X, Y, Z: Double): TVector; overload;
    //function  OGLPanelPos(AV: TVector): TVector; overload;
    function  OGLPanelPos(Pos: TVector): TVector;
    function  RotAngle(PixelDis: Integer): Single;
    procedure SetRotMode(const Value: TRotMode);
    function  ScreenRelation: Single;
    function  ScreenAdjustHeight: Single;
    function  ScreenAdjustWidth: Single;
    function  VectorAdd2(V: array of TVector): TVector;
  public
    CameraMatrix: TMatrix;
    ProjMatrix: TMatrix;
    OGLControl: TControl;
    PickedColor: array[0..3] of GLubyte;
    constructor Create;
    //destructor Destroy; override;
    function  AdjustWidth: Double;
    function  AdjustHeight: Double;
    function  ScreenHeight: Single;
    function  ScreenWidth: Single;	
    procedure CM_Identity;
    procedure PM_Identity;

    property RotMode: TRotMode read FRotMode write SetRotMode;
    property PointofCamera: TVector read FPointofCamera write SetPointofCamera;
    property PointofView: TVector read FPointofView write SetPointofView;
    property PointofRotation: TVector read FPointofRotation write SetPointofRotation;
    property PointofOrgin: TVector read FPointofOrgin write SetPointofOrgin;
    property IsMovePoR: Boolean read FIsMovePoR write SetIsMovePoR;
    property IsScreenAdjust: Boolean read FIsScreenAdjust write SetIsScreenAdjust;
    property XVectors: TVector read GetXVectors write FXVectors;
    property YVectors: TVector read GetYVectors write FYVectors;
    property ZVectors: TVector read GetZVectors write FZVectors;
  end;

implementation

uses
  Dialogs, Forms, Messages;

{$R CamerAHCursor.res}

function RectWidth(const Rect: TRect): Integer;
begin
  Result := Rect.Right - Rect.Left;
end;

function RectHeight(const Rect: TRect): Integer;
begin
  Result := Rect.Bottom - Rect.Top;
end;

constructor TglCamerAHBasis.Create;
begin
  inherited;
  CM_Identity;
  PM_Identity;

  RotMode := rmNothing;
  EnableRotion := False;
  IsMovePoR := True;
  IsScreenAdjust := True;

  with screen do
  begin
    Cursors[crZoonIn]  := LoadCursor(HInstance, 'ZOOMIN'); //Zoom in
    Cursors[crZoomOut] := LoadCursor(HInstance, 'ZOOMOUT'); //Zoom out
    Cursors[crSelect]  := LoadCursor(HInstance, 'SELECT'); //Select
    Cursors[crRotate]  := LoadCursor(HInstance, 'ROTATE'); //Rotate
    Cursors[crMove]    := LoadCursor(HInstance, 'MOVE'); //Move
    Cursors[crMoveDepth]    := LoadCursor(HInstance, 'MOVEDEPTH'); //MOVEDEPTH
  end;
end;

procedure TglCamerAHBasis.CM_Identity;
begin
  CameraMatrix := IdentityHmgMatrix;
end;

procedure TglCamerAHBasis.PM_Identity;
begin
  ProjMatrix := IdentityHmgMatrix;
end;

procedure TglCamerAHBasis.CM_Multiply(M: TMatrix);
begin
  CameraMatrix := MatrixMultiply(M, CameraMatrix);
end;

procedure TglCamerAHBasis.Colorpicking(X, Y: Integer);
begin
  Colorpicking(X, Y, PickedColor[0], PickedColor[1], PickedColor[2], PickedColor[3]);
end;

procedure TglCamerAHBasis.Colorpicking(X, Y: Integer; var R, G, B, A: GLubyte);
var
  pixel: array[0..3] of GLubyte;
begin
	glReadPixels(X, Round(ScreenHeight - Y), 1,1, GL_BGRA, GL_UNSIGNED_BYTE, @pixel);
  A := pixel[3];
  R := pixel[2];
  G := pixel[1];
  B := pixel[0];
end;

function TglCamerAHBasis.DistanceCameraCenter: Single;
begin
  Result := VectorLength(VectorSubtract(PointofCamera, PointofView));
end;

function TglCamerAHBasis.DistanceCameraRotation: Single;
begin
  Result := VectorLength(VectorSubtract(PointofCamera, PointofRotation));
end;

function TglCamerAHBasis.MoveVector(orientation: TVector; PixelDis: Integer): TVector;
begin
  Result :=  VectorLerp(NullHmgVector, orientation, PixelDis * AdjustWidth);
end;

function TglCamerAHBasis.MoveVectorX(PixelDis: Integer): TVector;
begin
  Result := MoveVector(XVectors, PixelDis);
end;

function TglCamerAHBasis.MoveVectorY(PixelDis: Integer): TVector;
begin
  Result := MoveVector(YVectors, PixelDis);
end;

function TglCamerAHBasis.MoveVectorZ(PixelDis: Integer): TVector;
begin
  Result := MoveVector(ZVectors, PixelDis);
end;

function TglCamerAHBasis.RotAngle(PixelDis: Integer): Single;
begin
  Result := ArcSin( (PixelDis * AdjustWidth/ 2) / DistanceCameraRotation );
end;

function TglCamerAHBasis.AdjustHeight: Double;
begin
  Result := 1 /
    Abs(OGLPanelPos(VectorScale(VectorNormalize(YVectors), 1))[1] - OGLPanelPos(NullHmgVector)[1]);
end;

function TglCamerAHBasis.AdjustWidth: Double;
begin
  Result := 1 /
    Abs(OGLPanelPos(VectorScale(VectorNormalize(XVectors), 1))[0] - OGLPanelPos(NullHmgVector)[0]);
end;

function TglCamerAHBasis.OGLPos(X, Y, Z: Double): TVector;
var
  Pos: TVector;
  Y_new: Double;
  viewport: CamerAHTypes.TVector4i;
  projection,
  modelview,
  M: TMatrix;
  NotError: Boolean;
begin
  if (Y = 0) then
    Y := 1;

  glGetFloatv(GL_MODELVIEW_MATRIX, @modelview);
  glGetFloatv(GL_PROJECTION_MATRIX, @projection);
  glGetIntegerv(GL_VIEWPORT, @viewport);
  Y_new := viewport[3] - y;
  Pos := VectorMake(X, Y_new, Z);

  M := MatrixMultiply(modelview, projection);

  NotError := UnProject(Pos, M, viewport, Result);

  if not NotError then
    showMessage('Fehler bei UnProject');
end;

function TglCamerAHBasis.OGLPos(AV: TVector): TVector;
begin
  Result := OGLPos(Round(AV[0]), Round(AV[1]), Round(AV[2]));
end;


function TglCamerAHBasis.OGLPanelPos(Pos: TVector): TVector;
var
  viewport: CamerAHTypes.TVector4i;
  projection,
  modelview,
  M: TMatrix;
  NotError: Boolean;
begin
  glGetFloatv(GL_MODELVIEW_MATRIX, @modelview);
  glGetFloatv(GL_PROJECTION_MATRIX, @projection);
  glGetIntegerv(GL_VIEWPORT, @viewport);

  M := MatrixMultiply(modelview, projection);
  NotError := Project(Pos, M, viewport, Result);

  if not NotError then
    showMessage('Fehler bei Project');

end;

//function TglCamerAHBasis.OGLPanelPos(AV: TVector): TVector;
//begin
//  Result := OGLPanelPos(AV[0], AV[1], AV[2]);
//end;

function TglCamerAHBasis.ScreenRelation: Single;
begin
  Result := ScreenWidth / ScreenHeight;
end;

function TglCamerAHBasis.ScreenAdjustHeight: Single;
begin
  if (ScreenHeight > ScreenWidth) and IsScreenAdjust then
    Result := 1/ ScreenRelation
  else
    Result := 1;
end;

function TglCamerAHBasis.ScreenAdjustWidth: Single;
begin
  if (ScreenWidth > ScreenHeight) and IsScreenAdjust then
    Result := ScreenRelation
  else
    Result := 1;
end;

function TglCamerAHBasis.ScreenHeight: Single;
var
  viewport: TVector;
begin
  glGetFloatv(GL_VIEWPORT, @viewport);
  Result := viewport[3];
end;

function TglCamerAHBasis.ScreenWidth: Single;
var
  viewport: TVector;
begin
  glGetFloatv(GL_VIEWPORT, @viewport);
  Result := viewport[2];
end;

procedure TglCamerAHBasis.SetIsMovePoR(const Value: Boolean);
begin
  FIsMovePoR := Value;
end;

procedure TglCamerAHBasis.SetIsScreenAdjust(const Value: Boolean);
begin
  FIsScreenAdjust := Value;
end;

procedure TglCamerAHBasis.SetPointofCamera(const Value: TVector);
begin
  FPointofCamera := Value;
end;

procedure TglCamerAHBasis.SetPointofOrgin(const Value: TVector);
begin
  FPointofOrgin := Value;
end;

procedure TglCamerAHBasis.SetPointofRotation(const Value: TVector);
begin
  FPointofRotation := Value;
end;

procedure TglCamerAHBasis.SetPointofView(const Value: TVector);
begin
  FPointofView := Value;
end;

procedure TglCamerAHBasis.SetRotMode(const Value: TRotMode);
begin
  if (ProjectionMode = pjPerspectiv) and (Value = rmZoomPlus)  then
    Exit;

  FRotMode := Value;

  if Assigned(OGLControl) then
  with OGLControl do
  begin
    case FRotMode of
      rmFlexAxis,
      rmRotX, rmRotY, rmRotZ,
      rmRotCoordX, rmRotCoordY, rmRotCoordZ,
      rmArcBall:            Cursor := crRotate;
      rmMove:               Cursor := crMove;
      rmDepthMove:          Cursor := crMoveDepth;
      rmZoomPlus:           Cursor := crZoonIn;
      rmZoomMinus:          Cursor := crDefault;
      rmNothing:            Cursor := crDefault;
      rmSelect:             Cursor := crSelect;
    end;
  end;
end;

function TglCamerAHBasis.VectorAdd2(V: array of TVector): TVector;
var
  i: Integer;
begin
  Result := NullHmgVector;
  for i := 0 to Length(V) - 1 do
    Result := VectorAdd(Result, V[i]);
end;

function TglCamerAHBasis.GetXVectors: TVector;
var
  X, Y, Z, M: Single;
  Matrix: TMatrix;
begin
  glGetFloatv(GL_MODELVIEW_MATRIX, @Matrix);
  X := Matrix[0][0];
  Y := Matrix[1][0];
  Z := Matrix[2][0];
  M := Matrix[3][0];
  Result := VectorMake(X, Y, Z, M);
end;

function TglCamerAHBasis.GetYVectors: TVector;
var
  X, Y, Z, M: Single;
  Matrix: TMatrix;
begin
  glGetFloatv(GL_MODELVIEW_MATRIX, @Matrix);
  X := Matrix[0][1];
  Y := Matrix[1][1];
  Z := Matrix[2][1];
  M := Matrix[3][1];
  Result := VectorMake(X, Y, Z, M);
end;

function TglCamerAHBasis.GetZVectors: TVector;
var
  X, Y, Z, M: Single;
  Matrix: TMatrix;
begin
  glGetFloatv(GL_MODELVIEW_MATRIX, @Matrix);
  X := Matrix[0][2];
  Y := Matrix[1][2];
  Z := Matrix[2][2];
  M := Matrix[3][2];
  Result := VectorMake(X, Y, Z, M);
end;

end.
